import re
import nltk
import numpy as np
import string
import requests
import networkx as nx
from enum import Enum
from nltk.corpus import stopwords
from nltk.cluster.util import cosine_distance
from collections import Counter
from morfeusz2 import Morfeusz
from bs4 import BeautifulSoup
from sklearn.feature_extraction.text import TfidfVectorizer

nltk.download("stopwords")
nltk.download('punkt_tab')

class KernelType(Enum):
    BOW = "bow"
    TFIDF = "tfidf"

class SummarizerException(Exception):
    pass

class Summarizer:

    def __init__(self, url_path: string, kernel: string, top_n: int) -> None:
        self.url_path = url_path
        if kernel not in [KernelType.BOW.value, KernelType.TFIDF.value]:
            raise SummarizerException("Kernel must be either 'bow' or 'tfidf'")
        self.kernel = kernel
        self.top_n = top_n

    @staticmethod
    def get_polish_stopwords() -> list[string]:
        with open("polish_stopwords.txt", "r") as f:
            return f.read().splitlines()

    def scrape_text_from_url(self) -> tuple[string, string, list[string]]:
        response = requests.get(self.url_path)
        if response.status_code == 200:
            soup = BeautifulSoup(response.content, 'html.parser')
            h1_title = soup.find('h1', class_='title').get_text() if soup.find('h1', class_='title') else 'No title found'
            p_lead = soup.find('p', class_='lead').get_text() if soup.find('p', class_='lead') else 'No lead found'
            div_content_parts = soup.find('div', class_='contentparts')
            p_content_parts = []
            if div_content_parts:
                for p in div_content_parts.find_all('p', class_='contentpart--text'):
                    text = p.get_text()
                    if '\xa0' not in text and 'ZOBACZ WIDEO' not in text and not (p.find('a') and text == p.find('a').get_text()):
                        p_content_parts.append(text)
            return h1_title, p_lead, p_content_parts
        else:
            raise SummarizerException(f"Failed to fetch data from {self.url_path}")

    @staticmethod
    def find_first_match(word, analyses) -> string:
        for analysis in analyses:
            if analysis[2][0] == word:
                return analysis[2][1].split(":")[0]
        return word

    def preprocess_text(self, text, lemmatize=True) -> string:
        morf = Morfeusz()
        text = text.lower()
        text = re.sub(r"\s+", " ", text).strip()
        text = re.sub(r"[(),]", "", text).strip()

        if lemmatize:
            lemmatized = [self.find_first_match(word, morf.analyse(word)) for word in text.split()]
            return " ".join(lemmatized)
        return text


    def read_article(self) -> tuple[list[str], list[list[str]]]:
        res = self.scrape_text_from_url()
        raw_text = '. '.join((res[0], res[1], ' '.join(res[2])) )

        original_sentences = raw_text.split(". ")

        processed_sentences = [self.preprocess_text(sentence).split(" ") for sentence in original_sentences if sentence.strip()]

        return original_sentences, processed_sentences

    def sentence_similarity(self, sent1: string, sent2: string) -> float:
        stop_words = self.get_polish_stopwords()

        filtered_sent1 = [word.lower() for word in sent1 if word.lower() not in stop_words]
        filtered_sent2 = [word.lower() for word in sent2 if word.lower() not in stop_words]

        word_counts1 = Counter(filtered_sent1)
        word_counts2 = Counter(filtered_sent2)

        all_words = set(word_counts1.keys()).union(set(word_counts2.keys()))

        vector1 = [word_counts1.get(word, 0) for word in all_words]
        vector2 = [word_counts2.get(word, 0) for word in all_words]

        return 1 - cosine_distance(vector1, vector2)

    def similarity_matrix_bow(self, sentences: list[list[str]]):
        similarity_matrix = np.zeros((len(sentences), len(sentences)))

        for i in range(len(sentences)):
            for j in range(len(sentences)):
                if i != j:
                    similarity_matrix[i][j] = self.sentence_similarity(sentences[i], sentences[j])

        return similarity_matrix

    def similarity_matrix_tfidf(self, sentences: list[list[str]]) -> list[list[float]]:
        stop_words = self.get_polish_stopwords()

        sentence_texts = [" ".join(sentence) for sentence in sentences]

        tfidf_vectorizer = TfidfVectorizer(stop_words=stop_words)
        tfidf_matrix = tfidf_vectorizer.fit_transform(sentence_texts)

        similarity_matrix = (tfidf_matrix * tfidf_matrix.T).toarray()
        return similarity_matrix


    def run(self) -> string:
        original_sentences, processed_sentences = self.read_article()
        if kernel == 'tfidf':
            sentence_similarity_matrix = self.similarity_matrix_tfidf(processed_sentences)
        elif kernel == 'bow':
            sentence_similarity_matrix = self.similarity_matrix_bow(processed_sentences)
        else:
            raise Exception("Unknown kernel")

        sentence_similarity_graph = nx.from_numpy_array(sentence_similarity_matrix)
        scores = nx.pagerank(sentence_similarity_graph)

        ranked_sentences = sorted(((scores[i], i) for i in range(len(processed_sentences))), reverse=True)

        num_sentences = min(self.top_n, len(ranked_sentences))
        summarize_text = [original_sentences[ranked_sentences[i][1]] for i in range(num_sentences)]

        return ". ".join(summarize_text)


if __name__ == "__main__":
    url = 'https://sportowefakty.wp.pl/skoki-narciarskie/1161190/malysz-wskazal-nowego-lidera-polskiej-kadry-olbrzymi-potencjal'
    kernel: str = "bow"
    summarizer = Summarizer(url, kernel, 4)
    summarized_text = summarizer.run()
    print(summarized_text)

